#include <iostream>
#include <fstream>
#include <sstream>
#include <float.h>
#include "GL/glew.h"
#include "GL/glut.h"
#include <IL/il.h>
#include <IL/ilut.h>
#include "OBJModel.h"
#include "glutil.h"
#include <stdlib.h>

using namespace std;
using namespace chag;

OBJModel::OBJModel(void)
{
}

OBJModel::~OBJModel(void)
{
}

void OBJModel::load(std::string fileName)
{
	std::ifstream file;

  // ensure only one type of slashes...
  std::replace(fileName.begin(), fileName.end(), '\\', '/');

  file.open(fileName.c_str());
	if (!file)
	{
		cout << "Error in openening file '" << fileName << "'" << endl;
		exit(1);
	}
	cout << "Loading OBJ file: '" << fileName << "'..." << endl;
  std::string basePath = fileName.substr(0, fileName.find_last_of('/')) + "/";
	loadOBJ(file, basePath);
	cout << " verts: " << getNumVerts() << endl;
}



void OBJModel::setMaterialDiffuseTextureId(std::string matName, int textureId)
{
  m_materials[matName].diffuse_map_id = textureId;
}



size_t OBJModel::getNumVerts()
{
	size_t result = 0;
	for (size_t i = 0; i < m_chunks.size(); ++i)
	{
		Chunk &chunk = m_chunks[i];
		result += chunk.m_positions.size();
	}
	return result;
}


struct ObjTri
{
	int v[3];
	int t[3];
	int n[3];
};


void OBJModel::loadOBJ(ifstream &file, std::string basePath)
{
	std::vector<float3> positions;
	std::vector<float3> normals;
	std::vector<float2> uvs;
	std::vector<ObjTri> tris;
	std::vector<std::pair<std::string, size_t> > materialChunks;

	cout << "  Reading data..." << endl << flush;
	std::string lineread;
	while(std::getline(file, lineread)) // Read line by line
	{
		stringstream ss(lineread); 
		std::string firstword; 
		ss >> firstword;
		if(firstword == "mtllib")
		{
			std::string materialFile;
			ss >> materialFile;
			loadMaterials(basePath + materialFile, basePath);
		}
		else if(firstword == "usemtl")
		{
			std::string materialName;
			ss >> materialName;
			if (materialChunks.size() == 0 || (*materialChunks.rbegin()).first != materialName)
			{
				materialChunks.push_back(std::make_pair(materialName, tris.size()));
			}
		}
		else if(firstword == "v")
		{
			float3 p;
			ss >> p.x >> p.y >> p.z;
			positions.push_back(p);
		}
		else if(firstword == "vn")
		{
			float3 n;
			ss >> n.x >> n.y >> n.z;
			normals.push_back(n);
		}
		else if(firstword == "vt")
		{
			float2 uv;
			ss >> uv.x >> uv.y;
			uvs.push_back(uv);
		}
		else if(firstword == "f")
		{			
			ObjTri t;
			// try a quad...
			bool quad = false;
			int qv = 0;
			int qt = 0;
			int qn = 0;
			if (sscanf(lineread.c_str(), "f%*[ \t]%d/%d/%d %d/%d/%d %d/%d/%d %d/%d/%d", &t.v[0], &t.t[0], &t.n[0], &t.v[1], &t.t[1], &t.n[1], &t.v[2], &t.t[2], &t.n[2], &qv, &qt, &qn) == 12)
			{
				quad = true;
			}
			else if (sscanf(lineread.c_str(), "f%*[ \t]%d//%d %d//%d %d//%d %d//%d", &t.v[0], &t.n[0], &t.v[1], &t.n[1], &t.v[2], &t.n[2], &qv, &qn) == 8)
			{
				t.t[0] = 0; t.t[1] = 0; t.t[2] = 0;
				quad = true;
			}

			if (!quad)
			{
				int ret = sscanf(lineread.c_str(), "f%*[ \t]%d/%d/%d %d/%d/%d %d/%d/%d", &t.v[0], &t.t[0], &t.n[0], &t.v[1], &t.t[1], &t.n[1], &t.v[2], &t.t[2], &t.n[2]);
				if(ret != 9) 
				{
					sscanf(lineread.c_str(), "f%*[ \t]%d//%d %d//%d %d//%d", &t.v[0], &t.n[0], &t.v[1], &t.n[1], &t.v[2], &t.n[2]);
					t.t[0] = 0; t.t[1] = 0; t.t[2] = 0;
				}
			}
			//			if(ret != 6)
			//				sscanf(lineread.c_str(), "f%*[ \t]%d/%d %d/%d %d/%d", &t.v[0], &t.t[0], &t.v[1], &t.t[1], &t.v[2], &t.t[2]);
			for (int i = 0; i < 3; ++i)
			{
				t.n[i] -= 1;
				t.v[i] -= 1;
				t.t[i] -= 1;
			}

			tris.push_back(t);
			if (quad)
			{
				t.n[1] = t.n[2];
				t.t[1] = t.t[2];
				t.v[1] = t.v[2];
				t.n[2] = qn - 1;
				t.t[2] = qt - 1;
				t.v[2] = qv - 1;

				tris.push_back(t);
			}
		}
	}
	cout << "  done." << endl;

	cout << "  Shuffling..." << flush;
	// Reshuffle the normals and vertices to be unique.
	for (size_t i = 0; i < materialChunks.size(); ++i)
	{
		Chunk chunk;
		chunk.material = &m_materials[materialChunks[i].first];
		const size_t start = materialChunks[i].second;
		const size_t end = i + 1 < materialChunks.size() ? materialChunks[i + 1].second : tris.size();
		for (size_t index = start; index < end; ++index)
		{
			for (int j = 0; j < 3; ++j)
			{
				chunk.m_normals.push_back(normals[tris[index].n[j]]);
				chunk.m_positions.push_back(positions[tris[index].v[j]]);
				if(tris[index].t[j] != -1)
					chunk.m_uvs.push_back(uvs[tris[index].t[j]]);
			}
		}
#if 0
    printf("// mat %s\n", materialChunks[i].first.c_str());
    printf("float3 normals[] = {\n");
		for (size_t index = start; index < end; ++index)
		{
      printf("{ %f, %f, %f },\n", normals[tris[index].n[0]].x, normals[tris[index].n[0]].y, normals[tris[index].n[0]].z);
      printf("{ %f, %f, %f },\n", normals[tris[index].n[1]].x, normals[tris[index].n[1]].y, normals[tris[index].n[1]].z);
      printf("{ %f, %f, %f },\n", normals[tris[index].n[2]].x, normals[tris[index].n[2]].y, normals[tris[index].n[2]].z);
    }
    printf("};\n");

    
    printf("float3 positions[] = {\n");
		for (size_t index = start; index < end; ++index)
		{
      printf("{ %f, %f, %f },\n", positions[tris[index].v[0]].x, positions[tris[index].v[0]].y, positions[tris[index].v[0]].z);
      printf("{ %f, %f, %f },\n", positions[tris[index].v[1]].x, positions[tris[index].v[1]].y, positions[tris[index].v[1]].z);
      printf("{ %f, %f, %f },\n", positions[tris[index].v[2]].x, positions[tris[index].v[2]].y, positions[tris[index].v[2]].z);
    }
    printf("};\n");
    
    printf("float3 texCoords[] = {\n");
		for (size_t index = start; index < end; ++index)
		{
      printf("{ %f, %f },\n", uvs[tris[index].t[0]].x, uvs[tris[index].t[0]].y);
      printf("{ %f, %f },\n", uvs[tris[index].t[1]].x, uvs[tris[index].t[1]].y);
      printf("{ %f, %f },\n", uvs[tris[index].t[2]].x, uvs[tris[index].t[2]].y);
    }
    printf("};\n");
#endif
    m_chunks.push_back(chunk);
	}
	cout << "done." << endl;

	// lastly we could look out for duplicates and compact the array down again, if we would.

	// Now, create a Vertex Array Object per chunk and be done with it
	for (size_t i = 0; i < m_chunks.size(); ++i)
	{
		Chunk &chunk = m_chunks[i];
		glGenVertexArrays(1, &chunk.m_vaob); 
		glBindVertexArray(chunk.m_vaob);

		glGenBuffers(1, &chunk.m_positions_bo); 
		glBindBuffer(GL_ARRAY_BUFFER_ARB, chunk.m_positions_bo);
		glBufferData(GL_ARRAY_BUFFER_ARB, chunk.m_positions.size() * sizeof(float3), 
			&chunk.m_positions[0].x, GL_STATIC_DRAW); 
		glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);	
		glEnableVertexAttribArray(0);

		glGenBuffers(1, &chunk.m_normals_bo); 
		glBindBuffer(GL_ARRAY_BUFFER_ARB, chunk.m_normals_bo);
		glBufferData(GL_ARRAY_BUFFER_ARB, chunk.m_normals.size() * sizeof(float3), 
			&chunk.m_normals[0].x, GL_STATIC_DRAW); 
		glVertexAttribPointer(1, 3, GL_FLOAT, false, 0, 0);	
		glEnableVertexAttribArray(1);

		if(chunk.m_uvs.size() > 0){
			glGenBuffers(1, &chunk.m_uvs_bo); 
			glBindBuffer(GL_ARRAY_BUFFER_ARB, chunk.m_uvs_bo);
			glBufferData(GL_ARRAY_BUFFER_ARB, chunk.m_uvs.size() * sizeof(float2), 
				&chunk.m_uvs[0].x, GL_STATIC_DRAW); 
			glVertexAttribPointer(2, 2, GL_FLOAT, false, 0, 0);	
		}
		glEnableVertexAttribArray(2);
	}		
}



void OBJModel::render()
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	for (size_t i = 0; i < m_chunks.size(); ++i)
	{
		Chunk &chunk = m_chunks[i];
		if(chunk.material->diffuse_map_id != -1){
			glActiveTexture(GL_TEXTURE0);
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, chunk.material->diffuse_map_id);
		}

		GLint current_program = 0; 
		glGetIntegerv(GL_CURRENT_PROGRAM, &current_program);
		glUniform1i(glGetUniformLocation(current_program, "has_diffuse_texture"), chunk.material->diffuse_map_id != -1);
		glUniform3fv(glGetUniformLocation(current_program, "material_diffuse_color"), 1, &chunk.material->diffuseColor.x);
		glUniform3fv(glGetUniformLocation(current_program, "material_specular_color"), 1, &chunk.material->specularColor.x);
		glUniform3fv(glGetUniformLocation(current_program, "material_ambient_color"), 1, &chunk.material->ambientColor.x);
		glUniform3fv(glGetUniformLocation(current_program, "material_emissive_color"), 1, &chunk.material->emissiveColor.x);

		glBindVertexArray(chunk.m_vaob);
		glDrawArrays(GL_TRIANGLES, 0, (GLsizei)chunk.m_positions.size());
	}
	glPopAttrib();
}

void OBJModel::loadMaterials(std::string fileName, std::string basePath )
{
	ifstream file;
	file.open(fileName.c_str());
	if (!file)
	{
		cout << "Error in openening file";
		exit(1);
	}
	std::string currentMaterial("");

	std::string lineread;
	while(std::getline(file, lineread)) // Read line by line
	{
		//cout << lineread << endl; 
		stringstream ss(lineread); 
		std::string firstword; 
		ss >> firstword;
		if(firstword == "newmtl")
		{
			ss >> currentMaterial;
			Material m = { { 0.5f, 0.5f, 0.5f, 1.0f }, { 0.5f, 0.5f, 0.5f, 1.0f }, 
				{ 0.5f, 0.5f, 0.5f, 1.0f }, { 0.0f, 0.0f, 0.0f, 1.0f }, -1 };
			m_materials[currentMaterial] = m;
			m_materials[currentMaterial].diffuse_map_id = -1; 
		}
		else if(firstword == "Ka")
		{
			float4 color;
			ss >> color.x >> color.y >> color.z;
			m_materials[currentMaterial].ambientColor = color;
		}
		else if(firstword == "Kd")
		{
			float4 color;
			ss >> color.x >> color.y >> color.z;
			m_materials[currentMaterial].diffuseColor = color;
		}
		else if(firstword == "Ks")
		{
			float4 color;
			ss >> color.x >> color.y >> color.z;
			m_materials[currentMaterial].specularColor = color;
		}
		else if(firstword == "Ke")
		{
			float4 color;
			ss >> color.x >> color.y >> color.z;
			m_materials[currentMaterial].emissiveColor = color;
		}
		else if(firstword == "map_Kd")
		{
			std::string fileName; 
			ss >> fileName;
			m_materials[currentMaterial].diffuse_map_id = (int) loadTexture(basePath + fileName, basePath);
		}
	}
}

unsigned int OBJModel::loadTexture(std::string fileName, std::string /*basePath*/ )
{
	GLuint texid = ilutGLLoadImage(const_cast<char *>(fileName.c_str()));
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texid);
	glGenerateMipmap(GL_TEXTURE_2D);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 16);

	cout << "    Loaded texture '" << fileName << "', (" << ilGetInteger(IL_IMAGE_WIDTH) << "x" << ilGetInteger(IL_IMAGE_HEIGHT) << ")" << endl; 
	glBindTexture(GL_TEXTURE_2D, 0);
	return texid;
}
